#include "filmSim.hpp"
#include <cmath>
#include <math.h>

template <typename T>
void downscaleDivisible1D(const matrix<T> input,
                          matrix<T> &output,
                          const int scaleFactor,
                          const bool interleaved);
template <typename T>
void downscaleBilinear1D(const matrix<T> input,
                         matrix<T> &output,
                         const int start,
                         const int end,
                         const double scaleFactor,
                         const bool interleaved);



//Scales the input to the output to fit within the output sizes.
void downscale_and_crop(const matrix<float> input,
                        matrix<float> &output,
                        const int inputStartX,
                        const int inputStartY,
                        const int inputEndX,
                        const int inputEndY,
                        const int outputXSizeLimit,
                        const int outputYSizeLimit)
{
    const int inputXSize = inputEndX - inputStartX + 1;
    const int inputYSize = inputEndY - inputStartY + 1;

    //If the output size limit is bigger than the image, shrink it to fit.
    const int outputXSize = min(inputXSize, outputXSizeLimit);
    const int outputYSize = min(inputYSize, outputYSizeLimit);

    //Determine the split of the scaling between integer and bilinear scaling.
    //Integer is much faster, but we can only do integer multiples.
    //Bilinear can only do shrinks between 1 and 2.
    const double overallScaleFactor = max(double(inputXSize)/double(outputXSize),double(inputYSize)/double(outputYSize));
    const int integerScaleFactor = floor(overallScaleFactor);
    //double bilinearScaleFactor = overallScaleFactor/double(integerScaleFactor);

    //Downscale in one direction
    matrix<float> bilinearX;
    matrix<float> bothX;
    downscaleBilinear1D(input,bilinearX,inputStartX,inputEndX,overallScaleFactor,true);
    downscaleDivisible1D(bilinearX,bothX,integerScaleFactor,true);

    //Then transpose
    matrix<float> bothXTransposed;
    bothXTransposed.set_size(bothX.nc(),bothX.nr());
    bothX.transpose_to(bothXTransposed);

    //Then downscale in the other direction.
    matrix<float> bothXTransposedBilinearY;
    matrix<float> bothXTransposedBothY;
    downscaleBilinear1D(bothXTransposed,bothXTransposedBilinearY,inputStartY,inputEndY,overallScaleFactor,false);
    downscaleDivisible1D(bothXTransposedBilinearY,bothXTransposedBothY,integerScaleFactor,false);

    //Then transpose to the output.
    output.set_size(bothXTransposedBothY.nc(),bothXTransposedBothY.nr());
    bothXTransposedBothY.transpose_to(output);
    return;
}

//Scales the image such that the number of columns is redeuced by integer factor scaleFactor.
template <typename T>
void downscaleDivisible1D(const matrix<T> input,
                          matrix<T> &output,
                          const int scaleFactor,
                          const bool interleaved)
{
    const int inputNumRows = input.nr();
    const int inputNumCols = input.nc();
    const int outputNumRows = inputNumRows;
    //We must use the ceiling here in order to not undersize the output matrix.
    int outputNumCols;
    if (interleaved)
    {
        outputNumCols = 3*ceil(inputNumCols/(3*double(scaleFactor)));
    }
    else
    {
        outputNumCols = ceil(inputNumCols/double(scaleFactor));
    }
    output.set_size(outputNumRows,outputNumCols);

    if (interleaved)
    {
        #pragma omp parallel for shared(output)
        for (int i = 0; i < outputNumRows; i++)
            for (int j = 0; j < outputNumCols; j = j+3)
            {
                double sumR = 0;
                double sumG = 0;
                double sumB = 0;
                for (int k = j*scaleFactor; k < (j+3)*scaleFactor; k = k + 3 )
                {
                    sumR += input(i,k);
                    sumG += input(i,k+1);
                    sumB += input(i,k+2);
                }
                output(i,j)   = sumR/scaleFactor;
                output(i,j+1) = sumG/scaleFactor;
                output(i,j+2) = sumB/scaleFactor;
            }
    }
    else
    {
        #pragma omp parallel for shared(output)
        for (int i = 0; i < outputNumRows; i++)
            for (int j = 0; j < outputNumCols; j++)
            {
                double sum = 0;
                for (int k = j*scaleFactor; k < (j+1)*scaleFactor; k++)
                {
                    sum += input(i,k);
                }
                output(i,j) = sum/scaleFactor;
            }
    }
    return;
}

//Scales the image such that the number of columns is reduced by a scaling factor between 1 and 2.
//The scaling factor is computed from the overall scale factor such that it ends up an integer multiple
// of the desired final size.
template <typename T>
void downscaleBilinear1D(const matrix<T> input,
                         matrix<T> &output,
                         const int start,
                         const int end,
                         const double overallScaleFactor,
                         const bool interleaved)
{
    const int inputNumRows = input.nr();
    const int inputNumCols = end - start + 1;

    const int outputNumRows = inputNumRows;

    const int endNumCols = max(1.0,round(double(inputNumCols)/overallScaleFactor));
    //We need to make sure that it ends up being a whole multiple of the final size.
    const float newOverallScaleFactor = inputNumCols/double(endNumCols);
    //We assume that floor of newOverallScaleFactor is the same as the original.
    const int intScaleFactor = max(1.0f,floor(newOverallScaleFactor));
    //This is the scale factor used for bilinear.
    const float scaleFactor = newOverallScaleFactor/intScaleFactor;

    const int outputNumCols = endNumCols * intScaleFactor;

    if(interleaved)
        output.set_size(outputNumRows,outputNumCols*3);
    else
        output.set_size(outputNumRows,outputNumCols);

    #pragma omp parallel for shared(output)
    for (int i = 0; i < outputNumRows; i++)
    {
        for (int j = 0; j < outputNumCols; j++)
        {
            const double inputPoint = (double(j) + 0.5)*scaleFactor -0.5 + double(start);
            const int inputStart = floor(inputPoint);
            const int inputEnd = ceil(inputPoint);
            double notUsed;
            const double endWeight = modf(inputPoint, &notUsed);
            const double startWeight = 1 - endWeight;
            if (interleaved)
                for (int c = 0; c < 3; c++)
                {
                    output(i,3*j + c) = startWeight*double(input(i,3*inputStart + c)) + endWeight*double(input(i,3*inputEnd + c));
                }
            else
                output(i,j) = startWeight*double(input(i,inputStart)) + endWeight*double(input(i,inputEnd));
        }
    }

}
